#include <stdio.h>
#include <EzGraph.h>
#include "config.h"
#include "anime.h"
#include "obj_draw.h"
#include "select_stage.h"



void anime_init(void){
}

void reset_anime_sequence(anime_sequence_t *seq){
    anime_fragment_t *frag = seq->fragments;

    seq->current_fragment_ix = 0;
    seq->x = seq->base_x;
    seq->y = seq->base_y;
    while (frag->type != af_end){
        frag->ctr = 0;
        frag ++;
    }
}

image_id_t interpret_anime_sequence(anime_sequence_t *seq, anime_mode_t mode){
    anime_fragment_t *frag;
    image_id_t iid;

    while (1){
        frag = &seq->fragments[seq->current_fragment_ix];
        //        printf("frag type:%d base (%d, %d) pos(%d, %d) iid:%d ctr:%d %p\n", frag->type, seq->base_x, seq->base_y, seq->x, seq->y, AnimeParamImgIid(frag), frag->ctr, frag);
        switch(frag->type){
        case af_img_static:
            if (frag->ctr == 0){
                seq->x += AnimeParamImgOffsetX(frag);
                seq->y += AnimeParamImgOffsetY(frag);
            }
            iid = AnimeParamImgIid(frag);
            //            if (AnimeParamImgIid(frag) != ii_none){
            //                draw_img(AnimeParamImgIid(frag), seq->x, seq->y);
            //            }
            frag->ctr ++;
            if (frag->ctr >= AnimeParamImgNum(frag) * ANIME_SLOWDOWN){
                frag->ctr = 0;
                seq->current_fragment_ix ++;
            }
            return iid;
        case af_loop:
			frag->ctr ++;
			if (frag->ctr <= AnimeParamLoopNum(frag)){
	            seq->current_fragment_ix += AnimeParamLoopOffset(frag);
			}
			else {
                frag->ctr = 0;
				seq->current_fragment_ix ++;
			}
            break;
        case af_end:
            reset_anime_sequence(seq);
            if (mode == am_single){
                return ii_invalid;
            }
            break;
        default:
            printf("anime default:\n");
            return ii_invalid;
        }
    }
}

int draw_anime_sequence(anime_sequence_t *seq, anime_mode_t mode){
    image_id_t iid;
    iid = interpret_anime_sequence(seq, mode);

    if (iid == ii_invalid){
        if (mode == am_single) return 1;
        // リピートモードなら、リセット直後のseqを使って先頭フレームを描画させる
        return draw_anime_sequence(seq, mode);
    }
    if (iid != ii_none){
        //        printf("draw anime %d\n", iid);
        draw_img(iid, seq->x, seq->y);
    }
    return 0;
}

// アニメーションをstepフレーム分すすめる。
// stepフレーム以内でaf_endに到達したら、非0を返す。
// それ以外は0を返す
int skip_anime(anime_sequence_t *seq, int step){
    int i;
    image_id_t iid;        
    for (i = 0 ; i < step ; i++){
        // skip時はリピートしないのでam_singleを指定
        iid = interpret_anime_sequence(seq, am_single);
        if (iid == ii_invalid) return 1;
    }
    return 0;
}


int draw_anime_array(anime_sequence_t *seq[], anime_mode_t mode){
    int no = 0, ret;

    while (seq[no] != NULL){
        ret = draw_anime_sequence(seq[no], mode);
        no++;
    }
    return ret;
}

void reset_anime_array(anime_sequence_t *seq[]){
    int no = 0;

    while (seq[no] != NULL){
        reset_anime_sequence(seq[no]);
        no++;
    }
}

int skip_anime_array(anime_sequence_t *seq[], int step){
    int no = 0, ret;

    while (seq[no] != NULL){
        ret = skip_anime(seq[no], step);
        no++;
    }
    return ret;
}


#ifdef LOCK_STAGE
int is_anime_locked(anime_sequence_t *seq){
    if (is_obj_introduced(seq->obj_type)) return 0;
    return 1;
}
int is_anime_array_locked(anime_sequence_t *seq[]){
    int no = 0;

    while (seq[no] != NULL){
        if (!is_obj_introduced(seq[no]->obj_type)) return 1;
        no++;
    }
    return 0;
}
#endif
