#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <EzGraph.h>
#include "config.h"
#include "layout.h"
#include "draw.h"
#include "game.h"
#include "obj_draw.h"
#include "title_draw.h"
#include "howtoplay_draw.h"
#include "select_stage_draw.h"
#include "stage.h"
#include "stage_info_draw.h"
#include "gameclear_draw.h"
#include "gameover_draw.h"
#include "font.h"
#include "game.h"
#include "obj.h"

#include "xpm/titles.xpm"

rgb_t colors[] = {
    {  0,   0,   0},
    {  0,   0, 255},
    {255,   0,   0},
    {255,   0, 255},
    {  0, 255,   0},
    {  0, 255, 255},
    {255, 255,   0},
    {255, 255, 255},
};

font_id_t msg_score[] = {
    fi_to, fi_ku, fi_te, fi_n, fi_end
};
font_id_t msg_left[] = {
    fi_no, fi_ko, fi_ri, fi_end
};

static EzImage title_img;

static char *gondin_str = "ごんでぃん";
static font_id_t *gondin_fontseq;
static char *gondin_toeat_str = "のるま";
static char *gondin_eaten_str = "たべた";
//  後半に数値を入れるので、明示的に確保
static font_id_t gondin_toeat_fontseq[10];
static font_id_t gondin_eaten_fontseq[10];
static char *stage_str = "すてーじ";
// stageは後半にステージ番号を入れるので、明示的に確保
static font_id_t stage_fontseq[10];

//static char *score_str = "とくてん";
//tatic font_id_t *score_fontseq;
//static font_id_t score_val_fontseq[10];
static char *time_str = "じかん";
static font_id_t *time_fontseq;
static font_id_t time_val_fontseq[GAME_TIME_LEN+1];

//static char *pause_str = "[[いちじていし]";
//static font_id_t *pause_fontseq;
static char *pause_cont_str = "[KeySpaceLeft][KeySpaceRight]さいかい";
static font_id_t *pause_cont_fontseq;
static char *pause_cont_init_str = "[KeySpaceLeft][KeySpaceRight]かいし";
static font_id_t *pause_cont_init_fontseq;
static char *pause_quit_str = "[KeyQ]すてーじせんたく";
static font_id_t *pause_quit_fontseq;
static char *pause_objective_str = "[KeyR]もくひょうをみる";
static font_id_t *pause_objective_fontseq;
static char *objective_str = "もくひょう";
static font_id_t *objective_fontseq;
static font_id_t objective_info_fontseqs[OBJECTIVE_NUM][10];
static font_id_t *objective_description_fontseqs[OBJECTIVE_NUM];

#define SLtx(x) ((x) + STATUS_XOFFSET)
#define SLty(y) ((y) + STATUS_YOFFSET)
#define SRbx(x) ((x) + WIN_WIDTH-1)
#define SRby(y) ((y) + WIN_HEIGHT - FONTSIZE-4)

static int gameclear_onoff = 0;
static int gameover_onoff = 0;
static unsigned int prev_time;
static int is_initial_pause = 0;
static int is_extend_objective = 0;

void draw_stage_result(void);
void draw_time(void);

void draw_init(void){
	obj_draw_init();
    font_init();
    title_draw_init();
    select_stage_draw_init();
    stage_info_draw_init();
    howtoplay_draw_init();
    gameclear_draw_init();
	gameover_draw_init();
    //    pause_fontseq = alloc_fontseq(pause_str);
    pause_cont_fontseq = alloc_fontseq(pause_cont_str);
    pause_cont_init_fontseq = alloc_fontseq(pause_cont_init_str);

	pause_quit_fontseq = alloc_fontseq(pause_quit_str);
	pause_objective_fontseq = alloc_fontseq(pause_objective_str);
    gondin_fontseq = alloc_fontseq(gondin_str);
	str2fontseq(stage_str, stage_fontseq);
	str2fontseq(gondin_toeat_str, gondin_toeat_fontseq);
	str2fontseq(gondin_eaten_str, gondin_eaten_fontseq);
//    score_fontseq = alloc_fontseq(score_str);
    time_fontseq = alloc_fontseq(time_str);
	objective_fontseq = alloc_fontseq(objective_str);
    title_img = EzReadXPM(title_s_xpm);
}

void setup_draw_screen(void){
    int i;
    objective_t *objective;

	uint2fontseq(get_current_stage_gondin_toeat(), 5, gondin_toeat_fontseq+3);
	uint2fontseq(0, 5, gondin_eaten_fontseq+3);
	uint2fontseq(get_current_stage_no(), 5, stage_fontseq+4);
	uint2fontseq(GetTime(), GAME_TIME_LEN, time_val_fontseq);
    for (i = 0 ; i < OBJECTIVE_NUM ; i++){
        objective = get_stage_objective(i);
        if (objective->entry != re_invalid){
            objective_info_fontseqs[i][0]
                = objective->icon[0];
            objective_info_fontseqs[i][1] = fi_space;
            switch (objective->op){
            case oo_le:
                objective_info_fontseqs[i][2] = fi_op_le;
                break;
            case oo_ge:
                objective_info_fontseqs[i][2] = fi_op_ge;
                break;
            case oo_eq:
                objective_info_fontseqs[i][2] = fi_equal;
                break;
            default:
                break;
            }
            objective_info_fontseqs[i][3] = fi_space;
            uint2fontseq(objective->val, 4, &objective_info_fontseqs[i][4]);
            objective_description_fontseqs[i] = 
                objective->description_fontseq;
        }
        else {
            objective_info_fontseqs[i][0] = fi_end;
        }
    }

    switch_draw_gameclear(0);
    switch_draw_gameover(0);
	prev_time = GetTime();
    is_initial_pause = 1;
    is_extend_objective = 0;
}

void switch_draw_gameclear(int onoff){
    gameclear_onoff = onoff;
}
void switch_draw_gameover(int onoff){
    gameover_onoff = onoff;
}

void draw_screen(void){
	int i;
  
    EzPutB(SLtx(0)-16, 4, title_img);

    draw_fontseq(SLtx(0), SLty(0), stage_fontseq);
    for (i = 0 ; i < STAGE_TITLE_ROW_NUM ; i++){
	    draw_fontseq(SLtx(16), SLty(LINESKIP*(i+1)),
                     get_current_stage_title(i));
    }
//    draw_fontseq(SLtx(0), SLty(LINESKIP*4), score_fontseq);
//	uint2fontseq(12234, 9, score_val_fontseq);
//    draw_fontseq(SLtx(0), SLty(LINESKIP*5), score_val_fontseq);
    draw_fontseq(SLtx(0), SLty(LINESKIP*3), gondin_fontseq);
//	uint2fontseq(12234, 9, score_val_fontseq);
    draw_fontseq(SLtx(16), SLty(LINESKIP*4), gondin_toeat_fontseq);
    draw_fontseq(SLtx(16), SLty(LINESKIP*5), gondin_eaten_fontseq);

    draw_fontseq(SLtx(0), SLty(LINESKIP*6), time_fontseq);
    if (prev_time != GetTime()){
        dec_fontseq(GAME_TIME_LEN, time_val_fontseq);
        prev_time = GetTime();
    }
    draw_fontseq(SLtx(0)+16*6, SLty(LINESKIP*7), time_val_fontseq);
    draw_fontseq(SLtx(0), SLty(LINESKIP*8), objective_fontseq);

    for (i = 0 ; i < OBJECTIVE_NUM ; i++){
		if (objective_info_fontseqs[i][0] != fi_end){
            if (is_extend_objective){
                draw_fontseq(SLtx(0), SLty(LINESKIP*(OBJ_YOFFSET+i*2)), 
                             game.objectives[i].stat_fontseq);
                draw_fontseq(SLtx(0)+16, SLty(LINESKIP*(OBJ_YOFFSET+i*2)), 
                             objective_info_fontseqs[i]);
                draw_fontseq(SLtx(0)+16, SLty(LINESKIP*(OBJ_YOFFSET+i*2+1)), 
                             objective_description_fontseqs[i]);
            }
            else {
                draw_fontseq(SLtx(0), SLty(LINESKIP*(OBJ_YOFFSET+i)), 
                             game.objectives[i].stat_fontseq);
                draw_fontseq(SLtx(0)+16, SLty(LINESKIP*(OBJ_YOFFSET+i)), 
                             objective_info_fontseqs[i]);
            }
		}
    }


//    draw_fontseq(SLtx(0), SLty(LINESKIP*8), gondin_fontseq);

    draw_sign_message();

    draw_obj_list();
#ifdef DEBUG
	if (GetShowHitMarker()){
        draw_hitmarker();
    }
#endif
    if (gameclear_onoff){
        draw_gameclear();
    }
    else if (gameover_onoff){
        draw_gameover();
    }
//    draw_time();
    EzShowBuffer();
}

void draw_sign_message(void){
    int i;
    int blx, bty, brx, bby, row_num;
    row_num = get_current_stage_sign_row_num();
    //    printf("row_num %d\n", row_num);
    if (get_current_stage_sign_no() != INVALID_LINK_NO){
        for (i = 0 ; i < STAGE_SIGN_ROW_NUM ; i++){
            draw_fontseq(SLtx(FONTSIZE)-FONTSIZE/2, 
                         SLty(LINESKIP*(i+SIGN_OFFSET
                                        +(is_extend_objective ? 3 : 0)))
                         +4,
			             get_current_stage_sign(i));
		}
        SetColor(7);
        blx = SLtx(0);
        bty = SLty(LINESKIP*(SIGN_OFFSET+(is_extend_objective ? 3 : 0)));
        brx = blx + FONTSIZE*(8+1);
        bby = bty + LINESKIP*row_num+FONTSIZE/4;            
        EzDrawBoxB(blx-1, bty, brx+1, bby);
        EzDrawBoxB(blx, bty-1, brx, bby+1);
	}
}



void draw_pause_message(void){
    if (is_extend_objective){
        //        draw_fontseq(SLtx(0), SRby(LINESKIP*19), 
        draw_fontseq(SLtx(0), SRby(0),
                     is_initial_pause ?
                     pause_cont_init_fontseq : pause_cont_fontseq);
    }
    else {
        //        draw_fontseq(SLtx(0), SLty(LINESKIP*17), 
        draw_fontseq(SLtx(0), SRby(-LINESKIP*2), 
                     is_initial_pause ?
                     pause_cont_init_fontseq : pause_cont_fontseq);
        draw_fontseq(SLtx(0), SRby(-LINESKIP), pause_quit_fontseq);
        draw_fontseq(SLtx(0), SRby(0), pause_objective_fontseq);
    }
}


void draw_time(void){
    char buf[12];

    
    sprintf(buf, "%2dfps", GetFramesPerSec());
    SetColor(7);
    EzDrawStringB(WIN_WIDTH - 8*5, 16, buf);
    //    puts(buf);
}

void inc_gondin_eaten_fontseq(void){
  inc_fontseq(5, gondin_eaten_fontseq+3);
}

void reset_initial_pause(void){
    is_initial_pause = 0;
}
int get_extend_objective(void){
    return is_extend_objective;
}
void set_extend_objective(int n){
    is_extend_objective = n;
}
