#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"
#include "emap.h"
#include "layout.h"
#include "uty.h"
#include "obj.h"

emap_e_t emap_static[MAP_HEIGHT][MAP_WIDTH];
emap_e_t emap[2][MAP_HEIGHT][MAP_WIDTH];
obj_t *emap_owner_static[en_ownered_end][MAP_HEIGHT][MAP_WIDTH];
obj_t *emap_owner[2][en_ownered_end][MAP_HEIGHT][MAP_WIDTH];

int current_emap = 0;

void emap_init(void){
    // ビットフィールドが32ビットに納まっているかチェック
    assert(en_end < 32);
}

void clear_emap(void){
    int x, y;
    memset(emap_static, 0, sizeof(emap_static));
    memset(emap, 0, sizeof(emap));
    memset(emap_owner_static, 0, sizeof(emap_owner_static));
    memset(emap_owner, 0, sizeof(emap_owner));

    // 周囲を壁で囲まれていないマップで、外に飛び出せないよう「壁」を作る
    for (x = 0 ; x < MAP_WIDTH ; x++){
        SetEMapStatic(x, MAP_LTY, EM_BLOCK_UP);
        SetEMapStatic(x, MAP_RBY, EM_BLOCK_DOWN);
    }
    for (y = 0 ; y < MAP_HEIGHT ; y++){
        SetEMapStatic(MAP_LTX, y, EM_BLOCK_LEFT);
        SetEMapStatic(MAP_RBX, y, EM_BLOCK_RIGHT);
    }
}

void clear_current_emap_owner(void){
    memset(emap_owner[current_emap], 0, sizeof(emap_owner)/2);
}

void setup_emap(void){
	switch_current_emap();
}

void switch_current_emap(void){
    clear_current_emap_owner();
	current_emap = NextEMap;
    cp_static_emap();
}

void copy_emap(void){
    memcpy(emap[current_emap], emap[NextEMap], 
           sizeof(emap_static));	
}

void cp_static_emap(void){
    memcpy(emap[NextEMap], emap_static,
           sizeof(emap_static));
}
void cp_or_static_emap(void){
	char *np, *sp;
	int i;

	np = (char*)emap[NextEMap];
	sp = (char*)emap_static;
	for (i = 0 ; i < sizeof(emap_static) ; i++){
		*np |= *sp;
		np ++;
		sp ++;
	}
}

void set_emap_static_box(int ltx, int lty, int width, int height, int emask){
    int x, y;

    for (y = lty ; y < lty+height ; y++){
        for (x = ltx ; x < ltx+width ; x++){
            emap_static[y][x] |= emask;
        }
    }
}
void reset_emap_static_box(int ltx, int lty, int width, int height, int emask){
    int x, y;

    emask = ~emask;

    for (y = lty ; y < lty+height ; y++){
        for (x = ltx ; x < ltx+width ; x++){
            emap_static[y][x] &= emask;
        }
    }
}
void set_emap_rel(obj_t *obj, int rx, int ry, int emask){
    int x, y;

    if (obj->dir == di_left){
        x = Ltx(obj) + rx;
        y = Lty(obj) + ry;
    }
    else if (obj->dir == di_right){
        x = Rtx(obj) - rx;
        y = Lty(obj) + ry;
    }
    else {
        assert(0);
    }
    SetEMap(x, y, emask);
}
void set_emap_box(int ltx, int lty, int width, int height, int emask){
    int x, y;

    for (y = lty ; y < lty+height ; y++){
        for (x = ltx ; x < ltx+width ; x++){
            emap[NextEMap][y][x] |= emask;
        }
    }
}
void set_emap_box_rel(obj_t *obj, int rltx, int rlty, int rrbx, int rrby,
                      int emask)
{
    int ltx, lty, width, height;

    width = rrbx - rltx + 1;
    height = rrby - rlty + 1;
    if (obj->dir == di_left){
        ltx = Ltx(obj) + rltx;
        lty = Lty(obj) + rlty;
    }
    else if (obj->dir == di_right){
        ltx = Rtx(obj) - rrbx;
        lty = Lty(obj) + rlty;
    }
    else {
        assert(0);
    }
    set_emap_box(ltx, lty, width, height, emask);
}





void reset_emap_box(int ltx, int lty, int width, int height, int emask){
    int x, y;

    emask = ~emask;

    for (y = lty ; y < lty+height ; y++){
        for (x = ltx ; x < ltx+width ; x++){
            emap[NextEMap][y][x] &= emask;
        }
    }
}

void set_emap_owner_box(int ltx, int lty, int width, int height,
                        obj_t *obj, effect_no_t en)
{
    int x, y;
    int emask = 1 << en;
    assert(en < en_end);

    set_emap_box(ltx, lty, width, height, emask);
    for (y = lty ; y < lty+height ; y++){
        for (x = ltx ; x < ltx+width ; x++){
		    emap_owner[NextEMap][en][y][x] = obj;
        }
    }
}

void set_emap_owner_static_box(int ltx, int lty, int width, int height,
                               obj_t *obj, effect_no_t en)
{
    int x, y;
    int emask = 1 << en;
    assert(en < en_end);

    set_emap_static_box(ltx, lty, width, height, emask);
    for (y = lty ; y < lty+height ; y++){
        for (x = ltx ; x < ltx+width ; x++){
		    emap_owner_static[en][y][x] = obj;
        }
    }
}

void set_emap_static_blockbox(int ltx, int lty, int width, int height){
    int x, y;
    // 基本的には外周部１ラインにブロック属性を設定するが、
    // すり抜け防止のため、物体の内側にも1ラインずつ設定しておく
	// マップ最外側のwallで範囲外アクセスが起きる可能性があるため、
	// 範囲チェックが必要。この関数が呼ばれるのはゲーム開始前のみなので
	// オーバヘッドは気にしなくて良い
    assert(ltx > 0);
    assert(ltx+width < MAP_WIDTH+MAP_MARGIN);
    assert(lty > 0);
    assert(lty+height < MAP_HEIGHT+MAP_MARGIN);
    for (x = ltx ; x < ltx + width ; x++){
        //        if (lty > 0) SetEMapStatic(x, lty-1, EM_BLOCK_DOWN);    
        SetEMapStatic(x, lty-1, EM_BLOCK_DOWN);    
        SetEMapStatic(x, lty, EM_EXCLUDE_UP);    
        SetEMapStatic(x, lty+1, EM_EXCLUDE_UP);    
        SetEMapStatic(x, lty+height-2, EM_EXCLUDE_DOWN);    
        SetEMapStatic(x, lty+height-1, EM_EXCLUDE_DOWN);    
        //        if (lty+height < MAP_HEIGHT) SetEMapStatic(x, lty+height, EM_BLOCK_UP);    
        SetEMapStatic(x, lty+height, EM_BLOCK_UP);    
    }
    for (y = lty ; y < lty + height ; y++){
        //        if (ltx > 0) SetEMapStatic(ltx-1, y, EM_BLOCK_RIGHT);    
        SetEMapStatic(ltx-1, y, EM_BLOCK_RIGHT);    
        SetEMapStatic(ltx, y, EM_EXCLUDE_LEFT);    
        SetEMapStatic(ltx+1, y, EM_EXCLUDE_LEFT);    
        SetEMapStatic(ltx+width-2, y, EM_EXCLUDE_RIGHT); 
        SetEMapStatic(ltx+width-1, y, EM_EXCLUDE_RIGHT); 
        //        if (ltx+width < MAP_WIDTH) SetEMapStatic(ltx+width, y, EM_BLOCK_LEFT); 
        SetEMapStatic(ltx+width, y, EM_BLOCK_LEFT); 
    }
}

void reset_emap_static_blockbox(int ltx, int lty, int width, int height){
    int x, y;
    // 基本的には外周部１ラインにブロック属性を設定するが、
    // すり抜け防止のため、物体の内側にも1ラインずつ設定しておく
    assert(ltx > 0);
    assert(ltx+width < MAP_WIDTH+MAP_MARGIN);
    assert(lty > 0);
    assert(lty+height < MAP_HEIGHT+MAP_MARGIN);
    for (x = ltx ; x < ltx + width ; x++){
        ResetEMapStatic(x, lty-1, EM_BLOCK_DOWN);    
        ResetEMapStatic(x, lty, EM_EXCLUDE_UP);
        ResetEMapStatic(x, lty+1, EM_EXCLUDE_UP);
        ResetEMapStatic(x, lty+height-2, EM_EXCLUDE_DOWN);
        ResetEMapStatic(x, lty+height-1, EM_EXCLUDE_DOWN);
        ResetEMapStatic(x, lty+height, EM_BLOCK_UP);    
    }
    for (y = lty ; y < lty + height ; y++){
        ResetEMapStatic(ltx-1, y, EM_BLOCK_RIGHT);    
        ResetEMapStatic(ltx, y, EM_EXCLUDE_LEFT);
        ResetEMapStatic(ltx+1, y, EM_EXCLUDE_LEFT);
        ResetEMapStatic(ltx+width-2, y, EM_EXCLUDE_RIGHT);
        ResetEMapStatic(ltx+width-1, y, EM_EXCLUDE_RIGHT);
        ResetEMapStatic(ltx+width, y, EM_BLOCK_LEFT); 
    }
}

void set_emap_blockbox(int ltx, int lty, int width, int height){
    int x, y;
    // 基本的には外周部１ラインにブロック属性を設定するが、
    // すり抜け防止のため、物体の内側にも1ラインずつ設定しておく
	// ※外周部分にスイッチなどを配置する場合に備え、
	// 　範囲外チェックを入れておく
	// 　実用上問題ないと思われるが、オーバヘッドが大きすぎるようなら、
	// 　emap配列に周囲1行・1列ずつのマージンを入れた方がいいかもしれない。
    assert(ltx > 0);
    assert(ltx+width < MAP_WIDTH+MAP_MARGIN);
    assert(lty > 0);
    assert(lty+height < MAP_HEIGHT+MAP_MARGIN);
    for (x = ltx ; x < ltx + width ; x++){
        //        if (lty > 0) SetEMap(x, lty-1, EM_BLOCK_DOWN);
        SetEMap(x, lty-1, EM_BLOCK_DOWN);
        SetEMap(x, lty, EM_EXCLUDE_UP);
        SetEMap(x, lty+1, EM_EXCLUDE_UP);
        SetEMap(x, lty+height-2, EM_EXCLUDE_DOWN);
        SetEMap(x, lty+height-1, EM_EXCLUDE_DOWN);
        //        if (lty+height < MAP_HEIGHT) SetEMap(x, lty+height, EM_BLOCK_UP);    
        SetEMap(x, lty+height, EM_BLOCK_UP);    
    }
    for (y = lty ; y < lty + height ; y++){
        //        if (ltx > 0) SetEMap(ltx-1, y, EM_BLOCK_RIGHT);
        SetEMap(ltx-1, y, EM_BLOCK_RIGHT);
        SetEMap(ltx, y, EM_EXCLUDE_LEFT);
        SetEMap(ltx+1, y, EM_EXCLUDE_LEFT);
        SetEMap(ltx+width-2, y, EM_EXCLUDE_RIGHT);
        SetEMap(ltx+width-1, y, EM_EXCLUDE_RIGHT);
        //        if (ltx+width < MAP_WIDTH) SetEMap(ltx+width, y, EM_BLOCK_LEFT); 
        SetEMap(ltx+width, y, EM_BLOCK_LEFT); 
    }
}

int check_emap_rel(obj_t *obj, int rx, int ry, int emask){
    int x, y;

    if (obj->dir == di_left){
        x = Ltx(obj) + rx;
        y = Lty(obj) + ry;
    }
    else if (obj->dir == di_right){
        x = Rtx(obj) - rx;
        y = Lty(obj) + ry;
    }
    else {
        assert(0);
    }

    return CheckEMap(x, y, emask);
}


int check_emap_box_any(int ltx, int lty, int width, int height, int emask){
	int x, y;

    for (y = lty ; y < lty+height ; y++){
	    for (x = ltx ; x < ltx+width ; x++){
            //			if (CheckEMap(x, y, eb)) return 1;
            ReturnCheckEMapPos(x, y, emask);
		}
	}
	return 0;
}
int check_emap_obj_any(obj_t *obj, int emask){
	return check_emap_box_any(obj->x, obj->y, obj->width, obj->height, emask);
}
int check_emap_pos_box_any(int ltx, int lty, int width, int height, int emask){
	int x, y;

    for (y = lty ; y < lty+height ; y++){
	    for (x = ltx ; x < ltx+width ; x++){
			if (CheckEMap(x, y, emask)){
                //                assert(GetEMapOwner(x, y, emask) != NULL);
                return Pos2CheckVal(x, y);
            }
		}
	}
	return 0;
}
int check_emap_pos_obj_any(obj_t *obj, int emask){
	return check_emap_pos_box_any(obj->x, obj->y, obj->width, obj->height,
                                  emask);
}

int check_emap_box_any_corner(int ltx, int lty, int width, int height,
                              int emask)
{
    if (CheckEMap(ltx, lty, emask) ||
        CheckEMap(ltx+width-1, lty, emask) ||
        CheckEMap(ltx, lty+height-1, emask) ||
        CheckEMap(ltx+width-1, lty+height-1, emask))
    {
        return 1;
    }
    else {
        return 0;
    }
}
int check_emap_obj_any_corner(obj_t *obj, int emask){
	return check_emap_box_any_corner(obj->x, obj->y, obj->width, obj->height,
                                     emask);
}
int check_emap_pos_box_any_corner(int ltx, int lty, int width, int height, 
                                  int emask)
{
	ReturnCheckEMapPos(ltx,         lty, emask);
	ReturnCheckEMapPos(ltx+width-1, lty, emask);
	ReturnCheckEMapPos(ltx,         lty+height-1, emask);
	ReturnCheckEMapPos(ltx+width-1, lty+height-1, emask);
    return 0;

}
int check_emap_pos_obj_any_corner(obj_t *obj, int emask){
	return check_emap_pos_box_any_corner(obj->x, obj->y,
                                         obj->width, obj->height, emask);
}

int check_emap_box_all_corner(int ltx, int lty, int width, int height,
                              int emask)
{
    if (CheckEMap(ltx, lty, emask) &&
        CheckEMap(ltx+width-1, lty, emask) &&
        CheckEMap(ltx, lty+height-1, emask) &&
        CheckEMap(ltx+width-1, lty+height-1, emask))
    {
        return 1;
    }
    else {
        return 0;
    }

}
int check_emap_obj_all_corner(obj_t *obj, int emask){
	return check_emap_box_all_corner(obj->x, obj->y, obj->width, obj->height,
                                     emask);
}

int check_emap_box_any_rim(int ltx, int lty, int width, int height, int emask){
    int v;
	if ((v = check_emap_box_any_left(ltx, lty, width, height, emask))) return v;
    if ((v = check_emap_box_any_right(ltx, lty, width, height, emask))) return v;
    if ((v = check_emap_box_any_up(ltx, lty, width, height, emask))) return v;
    if ((v = check_emap_box_any_down(ltx, lty, width, height, emask))) return v;
    return 0;
}
int check_emap_obj_any_rim(obj_t *obj, int emask){
	return check_emap_box_any_rim(obj->x, obj->y, obj->width, obj->height,
                                  emask);
}
// 相対位置指定のボックスに対してcheck_emap_box_any_rim
// ボックスはobjがdi_leftのときの(Ltx(obj), Lty(obj))に対する相対座標で
// ボックスの左上・右下隅座標を指定する
int check_emap_box_any_rim_rel(obj_t *obj, int rltx, int rlty,
                               int rrbx, int rrby, int emask)
{
    int ltx, lty, width, height;

    width = rrbx - rltx + 1;
    height = rrby - rlty + 1;
    if (obj->dir == di_left){
        ltx = Ltx(obj) + rltx;
        lty = Lty(obj) + rlty;
    }
    else if (obj->dir == di_right){
        ltx = Rtx(obj) - rrbx;
        lty = Lty(obj) + rlty;
    }
    else {
        assert(0);
    }
    return check_emap_box_any_rim(ltx, lty, width, height, emask);
}


int check_emap_box_any_left(int ltx, int lty, int width, int height, 
                            int emask)
{
    int i;
	for (i = 0 ; i < height ; i++){
        ReturnCheckEMapPos(ltx, lty+i, emask);
    }
    return 0;
}
int check_emap_box_any_right(int ltx, int lty, int width, int height,
                             int emask)
{
    int i;
	for (i = 0 ; i < height ; i++){
        ReturnCheckEMapPos(ltx+width-1, lty+i, emask);
    }
    return 0;
}
int check_emap_box_any_up(int ltx, int lty, int width, int height,
                          int emask)
{
    int i;
	for (i = 0 ; i < width ; i++){
        ReturnCheckEMapPos(ltx+i, lty, emask);
    }
    return 0;
}
int check_emap_box_any_down(int ltx, int lty, int width, int height,
                            int emask)
{
    int i;
	for (i = 0 ; i < width ; i++){
        ReturnCheckEMapPos(ltx+i, lty+height-1, emask);
    }
    return 0;
}
int check_emap_obj_any_left(obj_t *obj, int emask){
	return check_emap_box_any_left(obj->x, obj->y, obj->width, obj->height,
                                   emask);
}
int check_emap_obj_any_right(obj_t *obj, int emask){
	return check_emap_box_any_right(obj->x, obj->y, obj->width, obj->height,
                                    emask);
}
int check_emap_obj_any_up(obj_t *obj, int emask){
	return check_emap_box_any_up(obj->x, obj->y, obj->width, obj->height, 
                                 emask);
}
int check_emap_obj_any_down(obj_t *obj, int emask){
	return check_emap_box_any_down(obj->x, obj->y, obj->width, obj->height,
                                   emask);
}












int check_emap_dir(int emask, dir_t dir, obj_t *obj){
    switch(dir){
    case di_left:
        return check_emap_left(emask, obj);
    case di_right:
        return check_emap_right(emask, obj);
    case di_up:
        return check_emap_up(emask, obj);
    case di_down:
        return check_emap_down(emask, obj);
    default:
        break;
    }
    assert(0);
    exit(1);
}

int check_emap_left(int emask, obj_t *obj){
    int i, b=0;

    for (i = 0 ; i < obj->height ; i++){
        b |= CheckEMap(Ltx(obj), Lty(obj)+i, emask);
    }
    return b;
}
int check_emap_right(int emask, obj_t *obj){
    //    emap_e_t e;
    int i, b=0;

    for (i = 0 ; i < obj->height ; i++){
        b |= CheckEMap(Rtx(obj), Rty(obj)+i, emask);
        /*
        e = GetEMapE(Rtx(obj), Rty(obj)+i);
        b |= (check_emap_element(e, eb) ? (1 << i) : 0);
        */
    }
    return b;
}
int check_emap_up(int emask, obj_t *obj){
    //    emap_e_t e;
    int i, b=0;

    for (i = 0 ; i < obj->width ; i++){
        b |= CheckEMap(Ltx(obj)+i, Lty(obj), emask);
        /*
        e = GetEMapE(Ltx(obj)+i, Lty(obj));
        b |= (check_emap_element(e, eb) ? (1 << i) : 0);
        */
    }
    return b;
}
int check_emap_down(int emask, obj_t *obj){
    //    emap_e_t e;
    int i, b=0;

    for (i = 0 ; i < obj->width ; i++){
        b |= CheckEMap(Lbx(obj)+i, Lby(obj), emask);
        /*
        e = GetEMapE(Lbx(obj)+i, Lby(obj));
        b |= (check_emap_element(e, eb) ? (1 << i) : 0);
        */
    }
    return b;
}


int check_emap_any_rim(int emask, obj_t *obj){
    return check_emap_left(emask, obj) ||
		check_emap_right(emask, obj) ||
		check_emap_up(emask, obj) ||
		check_emap_down(emask, obj);
}





