#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
//#include "config.h"
#include "file.h"

// パス、ファイル名の最大長('\0'含まない)
// ディレクトリパスの最大長
#define MAX_DPATH 4096
// ファイル名の最大長
#define MAX_FILENAME 128
// ファイルパスの最大長(ディレクトリパス+'/'+ファイル名)
#define MAX_FPATH (MAX_DPATH+1+MAX_FILENAME)

// パス格納用バッファ(ファイルパス+'\0')
#define BUF_SIZE (MAX_FPATH+1)
#define LOG_OUTPUT_STDERR_DEFAULT 1
#define LOG_OUTPUT_FILE_DEFAULT 1
#define LOG_LIMIT_DEFAULT ((long)(1024*1024))

static char *exe_dpath;
// セーブディレクトリをカレントに仮設定
// @note file_init中でprint_logする可能性があるから、まずlog_fpathを仮設定する
static char save_dpath[MAX_DPATH+1] = "./";
static char log_filename[MAX_FILENAME+1] = "";
static char log_fpath[MAX_FPATH] = "";
static int log_output_stderr = LOG_OUTPUT_STDERR_DEFAULT;
static int log_output_file = LOG_OUTPUT_FILE_DEFAULT;
static long log_limit = LOG_LIMIT_DEFAULT;
static int enable_file = 0;


int file_init(char *argv0, char *dpath, char *save_dname, char *log_fname){
    char *home;
    char buf[BUF_SIZE];

    strcpy(log_filename, log_fname);
    // とりあえずlog_fpathを仮設定する
    set_log_fpath(log_filename);

    //    printf("dpath|%s| save_dname|%s|\n", dpath, save_dname);
    // セーブディレクトリを指定した場合
    if (dpath != NULL){
        //        printf("dpath:%s\n", dpath);
        return set_save_dpath(dpath);
    }
    home = getenv("HOME");
    // HOMEが存在する場合
    if (home != NULL){
        //        printf("home exist!\n");
        sprintf(buf, "%s/%s", home, save_dname);
        //        printf("|%s|\n", buf);
        return set_save_dpath(buf);
    }
    // 指定もHOMEもなければ実行ファイルのディレクトリをセーブディレクトリにする
    exe_dpath = get_exe_dpath(argv0);
    sprintf(buf, "%s", exe_dpath);
    return set_save_dpath(buf);
}

char *get_exe_dpath(char *argv0){
    static char dir[EXE_DIRNAME_MAX];
    char *sp = dir;
    char *ep;

    ep = strrchr(argv0, '/');
    // 絶対・相対パスつきで起動していたら、パス部分を返す
    // Win32版のargv[0]はフルパスなので、こちらで処理される
    if (ep != NULL){
        while (argv0 != ep+1){
            //            printf("%p %p %c\n", sp, ep, *argv0);
            *sp = *argv0;
            sp ++;
            argv0 ++;
        }
        *sp = '\0';
        return dir;
    }
    // 実行ファイル名のみの場合はPATHを順にチェック
    else {
        sp = getenv("PATH");
        ep = strchr(sp, ':');
        while (ep != NULL){
            strncpy(dir, sp, ep-sp);
            dir[ep-sp] = '/';
            strcpy(dir + (ep - sp + 1), argv0);
            //            printf("|%s|\n", dir);
            // 試したパスに実行ファイル名のファイルが実行属性付きであれば
            // それが答えと見なす
            if (access(dir, X_OK) == 0){
                dir[ep-sp+1] = '\0';
                return dir;
            }
            sp = ep + 1;
            ep = strchr(sp, ':');
        }
        return NULL;
    }
}

char *get_save_dpath(void){
    return save_dpath;
}

int set_save_dpath(char *dpath){
    strcpy(save_dpath, dpath);
    if (check_dir(save_dpath) == 0){
        log_warn("Cannot find '%s'.\n", save_dpath);
        if (make_dir(save_dpath) == 0){
            log_warn("Created '%s'.\n", save_dpath);
        }
        else {
            log_warn("Creating '%s' failed. Saving functions disabled.\n",
                     save_dpath);
            // 保存ディレクトリの作成に失敗したら、
            // 以後のデータload/saveは不可
            return 1;
        }
    }
    // ログの出力パスを新しいdpath下に再設定
    set_log_fpath(log_filename);
    enable_file = 1;
    return 0;
}

char *get_data_fpath(char *filename, char *fpath){
    sprintf(fpath, "%s/%s", save_dpath, filename);
    //	printf("open_data path:%s:\n", path);
	return fpath;
}

// SAVE_DIRにモードmodeでデータファイルfileを開く
// 
FILE *open_data(char *filename, char *mode,
                char *application_id, int application_id_size,
                unsigned char *major_version, unsigned char *minor_version)
{
    char fpath[BUF_SIZE];
    FILE *fp;
    unsigned char id[BUF_SIZE];

    if (!enable_file) return NULL;
    if (get_data_fpath(filename, fpath) == NULL) return NULL;
    fp = fopen(fpath, mode);
    if (fp == NULL) return NULL;

    // read openのとき
    if (strchr(mode, 'r')){
        // ヘッダが一致しなければ読み込み中止
        //        fread(id, 1, application_id_size, fp);
        checked_fread(id, application_id_size, fp, filename, "id");
        if (memcmp(id, application_id, application_id_size) != 0){
            fclose(fp);
            // このファイルは上書きされないようにリネームしておく
            rename_data(filename);
            return NULL;
        }
        // バージョン情報読み出し
        //        fread(major_version, sizeof(unsigned char), 1, fp);
        //        fread(minor_version, sizeof(unsigned char), 1, fp);
        checked_fread(major_version, sizeof(unsigned char), fp,
                    "major_version", filename);
        checked_fread(minor_version, sizeof(unsigned char), fp,
                    "minor_version", filename);
        //        printf("ver:%d %d\n", *major_version, *minor_version);
	}
    // write openのとき
    else {
        // ヘッダ書き込み
        fwrite(application_id, application_id_size, 1, fp);
        // バージョン情報書き込み
        fwrite(major_version, sizeof(unsigned char), 1, fp);
        fwrite(minor_version, sizeof(unsigned char), 1, fp);
        //        printf("ver:%d %d\n", *major_version, *minor_version);
    }
    return fp;
}

void checked_fread(unsigned char *buf, int size, FILE *fp,
                   char *filename, char *name)
{
    if (fread(buf, 1, size, fp) < size){
        if (name == NULL) name = "";
        if (filename == NULL) filename = "";
        log_error("Cannot read %s in %s", name, filename);
        exit(1);
    }
}

int fwrite_uint(FILE *fp, unsigned int *value, int num){
    int i, j;
    unsigned char c;

    for (i = 0 ; i < num ; i++){
        for (j = 3 ; j >= 0 ; j--){ // 上位バイトから順に出力
            c = (*value >> (8 * j)) & 0xff;
            fwrite(&c, sizeof(unsigned char), 1, fp);
        }
        value ++;
    }
    return 1;
}
int fread_uint(FILE *fp, unsigned int *value, int num,
               char *filename, char *name)
{
    int i, j;
    unsigned char c;

    for (i = 0 ; i < num ; i++){
        *value = 0;
        for (j = 3 ; j >= 0 ; j--){ // 上位バイトから順に入力
            //            fread(&c, sizeof(unsigned char), 1, fp);
            checked_fread(&c, sizeof(unsigned char), fp, filename, name);
            *value |= (c << (8 * j));
        }
        value ++;
    }
    return 1;
}


int set_log_fpath(char *log_filename){
	if (log_filename){
        sprintf(log_fpath, "%s/%s", save_dpath, log_filename);
	}
    else {
        log_fpath[0] = '\0';
    }
    return 0;
}

int print_log(const char *msg){
	FILE *fp;
    time_t current_time;
    char buf[BUF_SIZE];

    if (log_output_stderr){
        fputs(msg, stderr);
    }
    if (!log_output_file) return 1;
    if (log_fpath[0] == '\0') return 1;
    fp = fopen(log_fpath, "ab");
    if (!fp) return 1;
    // ログファイルのサイズがlog_limitを超えていたら、追加書き込みはやめる
    // @note バグにより無限ループ内でprint_logが呼ばれたとき、
    // ディスクをあふれさせるのを防ぐため
    //    printf("LOG LIMIT CHECK: ftell:%ld LOG_LIMIT:%ld\n", ftell(fp), log_limit);
    if (ftell(fp) > log_limit){
        // @todo あふれたら書き込み中止ではなく、ログをリネームしてリトライするよう
        //       変更
        fprintf(stderr, 
                "Error: The size of log file '%s' exceeds the limit "
                "(%ldMBytes).\n",
                log_fpath, log_limit/1024/1024);
        return 1;
    }
    fclose(fp);
    fp = fopen(log_fpath, "a");
    if (!fp) return 1;
    time(&current_time);
    strftime(buf, BUF_SIZE, "%Y-%m-%d %H:%M:%S ", localtime(&current_time));
    fputs(buf, fp);
    fputs(msg, fp);
    fclose(fp);
	return 0;
}

int log_warn(char *fmt, ...){
	va_list ap;
	char buf[BUF_SIZE] = "Warning: ";

	va_start(ap, fmt);
	vsprintf(buf+strlen(buf), fmt, ap);
    va_end(ap);
	return print_log(buf);
}

int log_error(char *fmt, ...){
	va_list ap;
	char buf[BUF_SIZE] = "Error: ";

	va_start(ap, fmt);
	vsprintf(buf+strlen(buf), fmt, ap);
    va_end(ap);
	return print_log(buf);
}


int make_dir(char *dpath){
    return mkdir(dpath, MAKEDIR_MODE);
}

int check_dir(char *dpath){
    struct stat buf;
    // stat()に失敗したらチェック失敗
    if (stat(dpath, &buf) != 0){
        return 0;
    }
    else {
        return S_ISDIR(buf.st_mode);
    }
}

int rename_data(char *filename){
    char fpath[BUF_SIZE];
    char buf[BUF_SIZE];
    int len, no = 0, i;

    get_data_fpath(filename, fpath);
    len = strlen(fpath);
    strcpy(buf, fpath);
    for (i = 0 ; i < 1000 ; i++){
        sprintf(buf+len, ".%03d", no);
        no++;
        if (rename(fpath, buf) == 0) return 0;
    }
    return -1;
} 


void set_log_output_stderr(int val){
    log_output_stderr = val;
}
void set_log_output_file(int val){
    log_output_file = val;
}
void set_log_limit(long val){
    log_limit = val;
}
