#ifndef FILE_H
#define FILE_H
#include <stdio.h>

#define EXE_DIRNAME_MAX 1024
#define MAKEDIR_MODE 0744

#define GetVersion(major, minor) (major*100+minor)

// argv[0]と、セーブディレクトリパス、
// HOME下のセーブディレクトリ名、ログファイル名を与える。
// セーブディレクトリを設定する
// - dpathがNULLでなければ、その値に設定
// - dpathがNULLなら、
//     - $HOMEが定義されていたら、その直下の(save_dirnameが示す名前の)ディレクトリ
//     - $HOMEが未定義なら、実行ファイルの存在するディレクトリ
// 前者の場合、ディレクトリがなければ作成する。
// セーブディレクトリの作成に失敗したら、本ファイルのセーブ・ロード機能は無効化される
// 正常に初期化できた場合は0、機能無効の場合は非0を返す
int file_init(char *argv0, char *dpath, char *save_dname, char *log_fname);

// 実行ファイルの存在するディレクトリパスを返す
char *get_exe_dpath(char *argv0);

// セーブディレクトリのパスを返す。
char *get_save_dpath(void);

// セーブディレクトリをdpathに変更する
// file_init時と同様、該当ディレクトリがなければ作成し、失敗したら機能無効化
// 正常に初期化できた場合は0、機能無効の場合は非0を返す
int set_save_dpath(char *dpath);

// データ保存ディレクトリにfilenameを保存するためのパスを作成しfpathに格納する
// パスの先頭へのポインタ(i.e. fpath)を返す
char *get_data_fpath(char *filename, char *fpath);

// SAVE_DIRにモードmodeでデータファイルfileを開く
// modeに'r'を含むとき、application_idが一致しなければファイルはクローズしNULLを返す
// (application_id==NULLなら一致チェックをしない)
// major_version / minor_versionsにはファイル内のバージョン値を格納する
FILE *open_data(char *filename, char *mode,
                char *application_id, int application_id_size,
                unsigned char *major_version, unsigned char *minor_version);

// エラーチェック・エラーログ出力つきのfread
// データ領域はunsigned charでsizeバイト分なので、通常のfreadに対し
//  sizeof(int), 4 => sizeof(int)*4
// などと置き換えが必要
// 読み込み失敗時にログにエラーを出力するため、
// 読み込み中のファイル名と読み込みデータ名(変数名など)を渡す
// 名前の記録が不要ならNULLを渡せばよい
void checked_fread(unsigned char *buf, int size, FILE *fp,
                   char *filename, char *name);

// unsigned intをネットワークバイトオーダでファイルから読み出す/に書き込む
// valueの指す先からnum個のunsigned int型を対象とする
// 読み書きは4バイト分で固定
// sizeof(unsigned int) > 4のときは下位4バイトのみが対象になる
// fread_uintの方は内部でchecked_freadを呼ぶため、データ名・ファイル名を引数にとる
int fwrite_uint(FILE *fp, unsigned int *value, int num);
int fread_uint(FILE *fp, unsigned int *value, int num,
               char *filename, char *name);

// ログファイルのパスを設定する。
// ※set_save_dirname()を先に呼ぶ必要がある。
// ※getarg.c:getarg()内で自動的に呼ばれる。
int set_log_fpath(char *log_filename);
int print_log(const char *msg);
int log_warn(char *fmt, ...);
int log_error(char *fmt, ...);

// dpathで示されるディレクトリを作成
// ※pathの最下層のディレクトリのみ作成する。
//   それより上のディレクトリが存在しなければ失敗する
// 成功時は0、失敗時は非0を返す
int make_dir(char *dpath);
// ディレクトリdpathが存在するなら非0を返す
int check_dir(char *dpath);
// pathをpath+".000", +".001", ...にリネームしようとする
// 成功したら0を返し、+".999"まですべて使用中なら失敗して-1を返す
int rename_data(char *filename);

void set_log_output_stderr(int val);
void set_log_output_file(int val);
void set_log_limit(long val);

#endif
