#include <stdio.h>
#include <EzGraph.h>
#include "config.h"
#include "layout.h"
#include "draw.h"
#include "obj_draw.h"
#include "stage.h"
#include "game.h"
#include "gameend.h"
#include "gameclear_draw.h"
#include "font.h"
#include "objective.h"

#define SCORE_LINELEN 24

static char *gameclear_str = "す　て　ー　じ　く　り　あ";
static font_id_t *gameclear_fontseq;
static font_id_t time_fontseq[SCORE_LINELEN+1];
static font_id_t gondin_fontseq[SCORE_LINELEN+1];
static font_id_t score_fontseq[SCORE_LINELEN+1];

static char *menu_strs[] = {
	"[KeyR] りとらい",
	"[KeyQ] おわる",
};
#define MENU_NUM (sizeof(menu_strs)/sizeof(char*))
static font_id_t **menu_fontseqs;
static int menu_x[MENU_NUM];


static font_id_t objective_info_fontseqs[OBJECTIVE_NUM][SCORE_LINELEN+1];

#define GCLtx(x) ((x) + GAMECLEAR_LTX)
#define GCLty(y) ((y) + GAMECLEAR_LTY)


void gameclear_draw_init(void){
	int i;
	assert(MENU_NUM == gi_end);
	gameclear_fontseq = alloc_fontseq(gameclear_str);
	menu_fontseqs = alloc_fontseq_array(MENU_NUM);
	for (i = 0 ; i < MENU_NUM ; i++){
		menu_fontseqs[i] = alloc_fontseq(menu_strs[i]);
		menu_x[i] = GAMECLEAR_LTX + 32 + i*((GAMECLEAR_WIDTH-32+32)/MENU_NUM);
	}
}

void setup_gameclear_draw(void){
    int i, j;
	char buf[128];
    objective_t *objective;

	sprintf(buf, "[Time] %3d*%3d=%8dてん",
            GetTime(), GAME_TIME_SCORE, GetScoreTime());
	str2fontseq(buf, time_fontseq);
	sprintf(buf, "[Gondin] %3d*%3d=%8dてん", 
            GetGondinEaten(), GONDIN_SCORE, GetScoreGondin());
	str2fontseq(buf, gondin_fontseq);

    for (i = 0 ; i < OBJECTIVE_NUM ; i++){
        objective = get_stage_objective(i);
        if (objective->entry != re_invalid){
			for (j = 0 ; j < SCORE_LINELEN ; j++){
				objective_info_fontseqs[i][j] = fi_space;
			}
			objective_info_fontseqs[i][0]
                = objective->icon[0];
			fontseq_cpy(&objective_info_fontseqs[i][1], objective->description_fontseq);
			objective_info_fontseqs[i][fontseq_len(objective->description_fontseq)+1] = fi_space;
            switch (objective->op){
            case oo_le:
                objective_info_fontseqs[i][9] = fi_op_le;
                break;
            case oo_ge:
                objective_info_fontseqs[i][9]= fi_op_ge;
                break;
            case oo_eq:
                objective_info_fontseqs[i][9] = fi_equal;
                break;
            default:
                break;
            }
            uint2fontseq(objective->val, 4, &objective_info_fontseqs[i][10]);
            sprintf(buf, "%4dてん", 
                    game.objectives[i].stat==st_done ? OBJECTIVE_SCORE : 0);
            str2fontseq(buf, &objective_info_fontseqs[i][14]);;
/*
            objective_info_fontseqs[i][1] = fi_space;
            switch (objective->op){
            case oo_le:
                objective_info_fontseqs[i][2] = fi_op_le;
                break;
            case oo_ge:
                objective_info_fontseqs[i][2] = fi_op_ge;
                break;
            case oo_eq:
                objective_info_fontseqs[i][2] = fi_equal;
                break;
            default:
                break;
            }
            objective_info_fontseqs[i][3] = fi_space;
            uint2fontseq(objective->val, 4, &objective_info_fontseqs[i][4]);
            sprintf(buf, "     %3dてん", 
                    game.objectives[i].stat==st_done ? OBJECTIVE_SCORE : 0);
            str2fontseq(buf, &objective_info_fontseqs[i][8]);;
*/
        }
        else {
            objective_info_fontseqs[i][0] = fi_end;
        }
    }




	sprintf(buf, "ごうけい　%6dてん", 
            GameRecord(re_score));
	str2fontseq(buf, score_fontseq);




//	uint2fontseq(playtime, GAME_TIME_LEN, time_fontseq+2);
//    time_fontseq[GAME_TIME_LEN+2] = fi_multiply; 
//	uint2fontseq(TIME_SCORE, 1, time_fontseq+4);
//	time_fontseq[GAME_TIME_LEN+4] = fi_equal; 
//	uint2fontseq(playtime*TIME_SCORE, GAME_TIME_LEN, time_fontseq+8);
}

void draw_gameclear(void){
	int i;
	gameend_menuitem_t cmd;

	cmd = gameend_cmd();

    SetColor(7);
    EzFillBoxB(GAMECLEAR_LTX-8, GAMECLEAR_LTY-8, GAMECLEAR_RBX+8, GAMECLEAR_RBY+8);
    SetColor(0);
    EzFillBoxB(GAMECLEAR_LTX, GAMECLEAR_LTY, GAMECLEAR_RBX, GAMECLEAR_RBY);
	draw_fontseq(GAMECLEAR_LTX+(GAMECLEAR_WIDTH-16*13)/2, GAMECLEAR_LTY+16, gameclear_fontseq);
	draw_fontseq(GAMECLEAR_LTX+16, GAMECLEAR_LTY+LINESKIP+24, time_fontseq);
	draw_fontseq(GAMECLEAR_LTX+16, GAMECLEAR_LTY+LINESKIP*2+24, gondin_fontseq);
    for (i = 0 ; i < OBJECTIVE_NUM ; i++){
        if (objective_info_fontseqs[i][0] != fi_end){
			draw_fontseq(GCLtx(0)+16, GCLty(LINESKIP*(4+i)), 
                         objective_info_fontseqs[i]);
		}
    }
	draw_fontseq(GAMECLEAR_LTX+16+16*7, GAMECLEAR_LTY+LINESKIP*6+24,
                 score_fontseq);
	for (i = 0 ; i < MENU_NUM ; i++){
		draw_fontseq(menu_x[i], GAMECLEAR_RBY-16-16, menu_fontseqs[i]);
	}
	draw_img(ii_chee16_wr0,
			 RDX(menu_x[cmd]-20), RDY(GAMECLEAR_RBY-16-12));
}

