#include <stdio.h>
#include <EzGraph.h>
#include "config.h"
#include "param.h"
#include "layout.h"
#include "howtoplay.h"
#include "howtoplay_draw.h"
#include "font.h"
#include "obj_draw.h"
#include "select_stage.h"
#include "draw.h"
#include "uty.h"

static char *title_str = "あ　そ　び　か　た";
static font_id_t *title_fontseq;
int title_x;

static char *menu_strs[] = {
	"るーる",
	"そうさほうほう",
    //	"とうじょうじん（？）ぶつ",
    "いきもの",
    "しかけ",
	"ちけい",
};
#define MENU_NUM (sizeof(menu_strs)/sizeof(char*))
static font_id_t **menu_fontseqs;
static int menu_x[MENU_NUM];



static char *rule_strs[] = {
	"ちぃちぃくん[Chee]を　そうさして　ごんでぃんのみ[Gondin]を　たべてください。",
	"「のるま」の　かずだけ　たべると　でぐち[Goal]が　ひらきます。",
	"でぐちに　はいると　すてーじくりあです。",
	"のこりじかんと　たべた　ごんでぃんのみが　とくてんに　なります。",
	"また、「もくひょう」の　じょうけんを　みたすと、"
	"とくてんが　ふえます。",
	"",
	"ひんと",
	"・ちぃちぃくんに　こうげきのうりょくは　ありません。",
	"　てきは　よけるか、ほかのてきや　ちけい・しかけを　つかって",
	"　たおしてください。",
	"・じかんが　0になっても　げーむおーばーには　なりません。",
//	"・じかんせいげんは　ありません。",
	"・まわりは　すべててきとは　かぎりません。",
	"　がいのないものや　りようできるものも　います。",
//	"　がいのないものや　じょうきょうにより　たすけになるものも　います。",
	"・たっせいした　もくひょうは　すてーじせんたくで　[StarOn]がひょうじされます。",
	"　いちどに　すべて　たっせいできるとは　かぎりません。",
	"　もくひょうに　あわせて　こうりゃくほうほうを　かえてみてください。",
};
#define RULE_NUM (sizeof(rule_strs)/sizeof(char*))
static font_id_t **rule_fontseqs;


key_info_t key_infos[] = {
	{"", "げーむ", "めにゅー", "すてーじせんたく", NULL, NULL, NULL, NULL},
	{"[KeyLeft]", "ひだりにいどう", "かーそるいどう", "かーそるいどう",
     NULL, NULL, NULL, NULL},
	{"[KeyRight]", "みぎにいどう", "かーそるいどう", "かーそるいどう",
     NULL, NULL, NULL, NULL},
	{"[KeyUp]", "じゃんぷ", "かーそるいどう", "かーそるいどう",
     NULL, NULL, NULL, NULL},
	{"[KeyDown]", NULL, "かーそるいどう", "かーそるいどう",
     NULL, NULL, NULL, NULL},
	{"[KeySpaceLeft][KeySpaceRight]", "じゃんぷ", "けってい", "げーむかいし", 
     NULL, NULL, NULL, NULL},
	{"[KeyEnter]", NULL, "けってい", "げーむかいし", 
     NULL, NULL, NULL, NULL},
	{"[KeyP]", "いちじていし", "", 
     NULL, NULL, NULL, NULL, NULL},
	{"[KeyEsc]", "いちじていし", "もどる", "もどる", 
     NULL, NULL, NULL, NULL},
	{"[KeyG]", "やりなおし", "", 
     NULL, NULL, NULL, NULL, NULL},
	{"[KeyQ]", "しゅうりょう", "もどる", "もどる", 
     NULL, NULL, NULL, NULL},
	{"[KeyR]", "もくひょうをみる", "", "きろくをみる", 
     NULL, NULL, NULL, NULL},
};
#define KEY_INFO_NUM (sizeof(key_infos)/sizeof(key_info_t))

object_info_t character_infos[] = {
    {ot_chee, "ちぃちぃくん", "ごんでぃんがすき", CHEE_SIZE, 
     {ii_chee_walk_l0, ii_chee_walk_l1, ii_end}, NULL,
     2, 0, 0, 
     NULL, NULL},
    {ot_gondin, "ごんでぃんのみ", "おいしい", GONDIN_SIZE,
     {ii_gondin, ii_end}, NULL,
     1, 0, 0,
     NULL, NULL},
    {ot_hatopo, "はとぽ", "じんちくむがい", HATOPO_SIZE,
     {ii_hatopo_walk_l0, ii_hatopo_walk_l1, ii_end},
     NULL,
     2, 0, 0,
     NULL, NULL},
    {ot_yaramasi, "やらまし", "とげがいたい", YARAMASI_HEIGHT, 
     {ii_yaramasi_walk_l0, ii_yaramasi_walk_l1, ii_end}, NULL,
     2, 0, 0, 
     NULL, NULL},
    {ot_mingo, "みんご", "てんてき", MINGO_SIZE, 
     {ii_mingo_walk_l0, ii_mingo_walk_l1, ii_mingo_walk_l2, ii_mingo_walk_l3,
//     {ii_mingo_attack_l0, ii_mingo_attack_l1,
      ii_end}, NULL,
     2, 0, 0, 
     NULL, NULL},
    {ot_gagga, "がっが", "つっこみちゅうい", GAGGA_SIZE,
     {ii_gagga_init, ii_end}, NULL,
     2, 0, 0, 
     NULL, NULL},
    {ot_ojigiri, "おじぎりそう", "いつもはらぺこ", OJIGIRI_HEIGHT,
     {ii_ojigiri_l, ii_ojigiri_r, ii_end}, NULL,
     8, 0, 0,
     NULL, NULL},
    {ot_sue, "すー", "ひをはく", SUE_HEIGHT,
     {ii_sue_high_close_l, ii_sue_high_open_l, ii_sue_low_close_l,
      ii_sue_low_open_l, ii_end}, NULL,
     16, 0, 0, 
     NULL, NULL},
    {ot_chu, "ちゅっちゅ", "ごんでぃんがすき", CHU_WALK_HEIGHT,
     {ii_chu_walk_l0, ii_chu_walk_l1, ii_end}, NULL,
     2, 0, 0,
     NULL, NULL},
    {ot_jumeme, "じゅめめ", "もやす・とかす", JUMEME_HEIGHT,
     {ii_jumeme_walk_l0, ii_jumeme_walk_l1, ii_end}, NULL,
     2, 0, 0,
     NULL, NULL},
    {ot_siguku, "しぐく", "こおる・ひをけす", SIGUKU_HEIGHT,
     {ii_siguku, ii_end}, NULL,
     4, 0, 0,
     NULL, NULL},
    {ot_ppit, "ぴぃひゃらぴっと", "とてもおくびょう", PPIT_HEIGHT,
     {ii_ppit_init, 
      ii_ppit_sl, ii_ppit_sl, ii_ppit_sl, 
      ii_ppit_init,
      ii_ppit_sr, ii_ppit_sr, ii_ppit_sr,
      ii_end}, NULL,
     4, 0, 0,
     NULL, NULL},
};
#define CHARACTER_INFO_NUM (sizeof(character_infos)/sizeof(object_info_t))
object_info_t gadget_infos[] = {
    {ot_mswitch, "たっちすいっち", "ふむとどうさ", GRID_SIZE,
     {ii_mswitch_off, ii_mswitch_on, ii_end}, NULL,
     20, 0, 0,
     NULL, NULL},
    {ot_aswitch, "とぐるすいっち", "ふむときりかわる", GRID_SIZE,
//     {ii_aswitch_off, ii_aswitch_on}, draw_gadget_aswitch,
     // ※自前描画なのでii_noneはmotionを0, 1と変化させるためのダミー
     {ii_none, ii_none, ii_end}, draw_gadget_aswitch,
     20, 0, 0,
     NULL, NULL},
    {ot_teleporter, "てれぽーた", "どこかへてんそう", GRID_SIZE,
     {ii_teleporter_off, ii_teleporter_on, ii_end}, NULL,
     20, 0, 0,
     NULL, NULL},
    {ot_lift, "りふと", "のっていどう", GRID_SIZE,
     {ii_none, ii_none, ii_end}, draw_gadget_lift,
     8, 0, 0,
     NULL, NULL},
    {ot_nya, "にゃあぱんち", "うでがのびーる", NYA_HEIGHT, 
     {ii_nya_on_l, ii_end}, NULL,
     2, 0, 0,
     NULL, NULL},
    {ot_pipe, "すいどうかん", "しぐくがおちる", GRID_SIZE, 
     {ii_pipe, ii_end}, NULL,
     0, 0, 0,
     NULL, NULL},
    {ot_barrier, "ばりあ", "とおれない", GRID_SIZE,
     {ii_barrier0, ii_barrier1, ii_barrier2, ii_barrier3, ii_end}, NULL,
     2, 0, 0,
     NULL, NULL},
    {ot_flame, "ほのお", "とてもあつい", GRID_SIZE, 
     {ii_flame_l, ii_end}, NULL,
     0, 0, 0,
     NULL, NULL},
};
#define GADGET_INFO_NUM (sizeof(gadget_infos)/sizeof(object_info_t))

object_info_t terran_infos[] = {
    {ot_wall, "かべ", "こわれない", GRID_SIZE,
     {ii_wall, ii_end}, NULL,
     0, 0, 0,
     NULL, NULL},
    {ot_fwall, "もろいかべ", "たたくとこわれる", GRID_SIZE,
     {ii_fwall, ii_end}, NULL,
     0, 0, 0,
     NULL, NULL},
    {ot_iwall, "こおりのかべ", "ほのおでとける", GRID_SIZE,
     {ii_iwall, ii_end}, NULL,
     0, 0, 0,
     NULL, NULL},
    {ot_stairs_l, "かいだん", "あるいてのぼれる", GRID_SIZE,
     {ii_stairs_l, ii_end}, NULL,
     0, 0, 0,
     NULL, NULL},
    {ot_slope_l, "しゃめん", "つるつるすべる", GRID_SIZE, 
     {ii_slope_l, ii_end}, NULL,
     0, 0, 0,
     NULL, NULL},
    {ot_sting, "とげ", "ささるといたい", GRID_SIZE,
     {ii_sting_u, ii_end}, NULL,
     0, 0, 0,
     NULL, NULL},
    {ot_sign, "たてふだ", "なにかかいてある", SIGN_HEIGHT, 
     {ii_sign, ii_sign_hang, ii_end}, NULL,
     50, 0, 0,
     NULL, NULL},
    {ot_goal, "でぐち", "はいるとくりあ", GOAL_HEIGHT,
     {ii_goal_closed, ii_end}, NULL,
     0, 0, 0, 
     NULL, NULL},
};

#define TERRAN_INFO_NUM (sizeof(terran_infos)/sizeof(object_info_t))


void howtoplay_draw_init(void){
    int i;
    title_fontseq = alloc_fontseq(title_str);
	title_x = FontSeqCenterLx(title_fontseq);
	menu_fontseqs = alloc_fontseq_array(MENU_NUM);
	for (i = 0 ; i < MENU_NUM ; i++){
		menu_fontseqs[i] = alloc_fontseq(menu_strs[i]);
		if (i == 0){
			menu_x[i] = 40;
		}
		else {
			menu_x[i] = menu_x[i-1]+FontSeqWidth(menu_fontseqs[i-1])+32;
		}
	}
	rule_fontseqs = alloc_fontseq_array(RULE_NUM);
	for (i = 0 ; i < RULE_NUM ; i++){
		rule_fontseqs[i] = alloc_fontseq(rule_strs[i]);
    }
	for (i = 0 ; i < KEY_INFO_NUM ; i++){
	    key_infos[i].key_fontseq
		    = alloc_fontseq(key_infos[i].key_str);
		if (key_infos[i].description_game_str != NULL){
			key_infos[i].description_game_fontseq 
				= alloc_fontseq(key_infos[i].description_game_str);
		}
		if (key_infos[i].description_menu_str != NULL){
			key_infos[i].description_menu_fontseq 
				= alloc_fontseq(key_infos[i].description_menu_str);
		}
		if (key_infos[i].description_stage_select_str != NULL){
			key_infos[i].description_stage_select_fontseq 
				= alloc_fontseq(key_infos[i].description_stage_select_str);
		}
	}
	for (i = 0 ; i < CHARACTER_INFO_NUM ; i++){
        character_infos[i].name_fontseq
            = alloc_fontseq(character_infos[i].name_str);
        character_infos[i].description_fontseq 
            = alloc_fontseq(character_infos[i].description_str);
	}
	for (i = 0 ; i < GADGET_INFO_NUM ; i++){
        gadget_infos[i].name_fontseq
            = alloc_fontseq(gadget_infos[i].name_str);
        gadget_infos[i].description_fontseq 
            = alloc_fontseq(gadget_infos[i].description_str);
	}
	for (i = 0 ; i < TERRAN_INFO_NUM ; i++){
        terran_infos[i].name_fontseq
            = alloc_fontseq(terran_infos[i].name_str);
        terran_infos[i].description_fontseq 
            = alloc_fontseq(terran_infos[i].description_str);
	}
}

void draw_howtoplay(void){
    int i;
	howtoplay_menuitem_t mode;

	mode = howtoplay_mode();
    draw_fontseq(title_x, 8, title_fontseq);

	for (i = 0 ; i < MENU_NUM ; i++){
        draw_fontseq(menu_x[i], 8+LINESKIP+8, menu_fontseqs[i]);
	}
//    draw_img(ii_punch_r,
    draw_img(ii_chee16_wr0,
			 RDX(menu_x[mode]-20), RDY(8+LINESKIP+8));

	switch(mode){
	case hi_rule:
		draw_rule();
		break;
	case hi_keys:
		draw_keys();
		break;
	case hi_character:
		draw_objects(CHARACTER_INFO_NUM, character_infos);
		break;
	case hi_gadget:
		draw_objects(GADGET_INFO_NUM, gadget_infos);
		break;
	case hi_terran:
		draw_objects(TERRAN_INFO_NUM, terran_infos);
		break;
	default:
		break;
	}
	EzShowBuffer();    
}
void draw_rule(void){
	int i;
	for (i = 0 ; i < RULE_NUM ; i++){
        draw_fontseq(32, HOWTOPLAY_VIEW_LTY+i*LINESKIP, rule_fontseqs[i]);
    }
}
void draw_keys(void){
	int i;
	for (i = 0 ; i < KEY_INFO_NUM ; i++){
        draw_fontseq(64,    HOWTOPLAY_VIEW_LTY+i*LINESKIP*3/2, key_infos[i].key_fontseq);
        if (key_infos[i].description_game_fontseq != NULL){
			draw_fontseq(64+48, HOWTOPLAY_VIEW_LTY+i*LINESKIP*3/2, key_infos[i].description_game_fontseq);
		}
        if (key_infos[i].description_menu_fontseq != NULL){
	        draw_fontseq(64+48+160, HOWTOPLAY_VIEW_LTY+i*LINESKIP*3/2, key_infos[i].description_menu_fontseq);
		}
        if (key_infos[i].description_stage_select_fontseq != NULL){
	        draw_fontseq(64+48+160*2, HOWTOPLAY_VIEW_LTY+i*LINESKIP*3/2, key_infos[i].description_stage_select_fontseq);
		}
    }
}


void draw_objects(int obj_num, object_info_t obj_infos[]){
	int i, x, y;
	
	x = HOWTOPLAY_VIEW_LTX;
	y = HOWTOPLAY_VIEW_LTY;



	for (i = 0 ; i < obj_num ; i++){
#ifdef LOCK_STAGE
        if (is_obj_introduced(obj_infos[i].type)){
#endif
        draw_fontseq(x+48+16, y, obj_infos[i].name_fontseq);
        draw_fontseq(x+48+16, y+LINESKIP, obj_infos[i].description_fontseq);
        obj_infos[i].motion_count ++;
        if (obj_infos[i].motion_count == 
            obj_infos[i].motion_interval)
        {
            obj_infos[i].motion ++;
            if (obj_infos[i].img[obj_infos[i].motion] == ii_end){
                obj_infos[i].motion = 0;
            }
            obj_infos[i].motion_count = 0;
        }
		if (obj_infos[i].img[0] != ii_none){
            draw_img(obj_infos[i].img[obj_infos[i].motion],
                     RDX(x),
                     RDY(y));
        }
		if (obj_infos[i].draw_func){
	        obj_infos[i].draw_func(RDX(x), RDY(y), &obj_infos[i]);
		}

#ifdef LOCK_STAGE
        }
        else {
            draw_font(x+48+16, y, fi_question);
            draw_font(x+48+16, y+LINESKIP, fi_question);
        }
#endif
		y += MaxInt(LINESKIP*3, DScale(obj_infos[i].height)+16);
		if (y > 400){
			x +=240;
			y = HOWTOPLAY_VIEW_LTY;
		}
    }
}

void draw_gadget_lift(int x, int y, object_info_t *objinfo){
	obj_t obj;
	obj.x = x;
    if (objinfo->motion == 0){
        obj.y = y + objinfo->motion_count;
    }
    else {
        obj.y = y+GRID_SIZE*2 - objinfo->motion_count;
    }
	obj.width = GRID_SIZE*2;
	obj.height = GRID_SIZE;
	draw_lift(&obj);
}

void draw_gadget_aswitch(int x, int y, object_info_t *objinfo){
	if (objinfo->motion == 0){
		draw_img(ii_aswitch_off, x, y);
	}
	else {
		draw_img(ii_aswitch_on, x, y+1);
	}
}
